from yta_editor.track.parts.abstract import _PartWithAudio, _PartWithVideo
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


class _VideoPart(_PartWithAudio, _PartWithVideo):
    """
    Class to represent an element that is on the
    track, that can be an empty space or a video.
    """

    def __init__(
        self,
        track: 'VideoTrack',
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction],
        media: Union['VideoOnTrack', None] = None
    ):
        ParameterValidator.validate_instance_of('media', media, 'VideoOnTrack')

        super().__init__(
            track = track,
            start = start,
            end = end,
            media = media
        )


"""
TODO: I don't like the name nor the
location of this file, but it is here
to encapsulate some functionality 
related to combining video frames.

Module to contain methods that combine
video frames. Call them with the 2
frames you want to combine and you 
will get the combined frame as return.
"""
from av.audio.resampler import AudioResampler
from av.audio.frame import AudioFrame

import numpy as np


# TODO: There was a 'VideoFrameCombinator'
# class here, but it is now in
# 'yta_numpy.rgba.combinator'

class AudioFrameCombinator:
    """
    Class to wrap the functionality related
    to combine different audio frames.
    """

    @staticmethod
    def sum_tracks_frames(
        tracks_frames: list[AudioFrame],
        sample_rate: int = 44100,
        layout: str = 'stereo',
        format: str = 'fltp',
        do_normalize: bool = True
    ) -> AudioFrame:
        """
        Sum all the audio frames from the different
        tracks that are given in the 'tracks_frames'
        list (each column is a single audio frame of
        a track). This must be a list that should 
        come from a converted matrix that was
        representing each track in a row and the
        different audio frames for that track on each
        column.

        This method is to sum audio frames of one
        specific 't' time moment of a video.

        The output will be the sum of all the audio
        frames and it will be normalized to avoid
        distortion if 'do_normalize' is True (it is
        recommended).
        """
        if len(tracks_frames) == 0:
            raise Exception('The "tracks_frames" list of audio frames is empty.')
        
        arrays = []
        resampler: AudioResampler = AudioResampler(
            format = format,
            layout = layout,
            rate = sample_rate
        )

        for track_frame in tracks_frames:
            # Resample to output format
            # TODO: What if the resampler creates more
            # than one single frame? I don't know what
            # to do... I'll see when it happens
            print(track_frame)
            track_frame = resampler.resample(track_frame)
            
            if len(track_frame) > 1:
                print('[ ! ]   The resampler has given more than 1 frame...')

            track_frame_array = track_frame[0].to_ndarray()

            # Transform to 'float32' [-1, 1]
            # TODO: I think this is because the output
            # is 'fltp' but we have more combinations
            # so this must be refactored
            if track_frame_array.dtype == np.int16:
                track_frame_array = track_frame_array.astype(np.float32) / 32768.0
            elif track_frame_array.dtype != np.float32:
                track_frame_array = track_frame_array.astype(np.float32)

            # Mono to stereo if needed
            # TODO: What if source is 'stereo' and we
            # want mono (?)
            if (
                track_frame_array.shape[0] == 1 and
                layout == 'stereo'
            ):
                track_frame_array = np.repeat(track_frame_array, 2, axis = 0)

            arrays.append(track_frame_array)

        # Same length and fill with zeros if needed
        max_len = max(a.shape[1] for a in arrays)
        stacked = []
        for a in arrays:
            # TODO: Again, this 'float32' is because output
            # is 'fltp' I think...
            buf = np.zeros((a.shape[0], max_len), dtype = np.float32)
            buf[:, :a.shape[1]] = a
            stacked.append(buf)

        """
        All this below is interesting to avoid
        distortion or things like that, but it
        is actually making the volume decrease
        in these parts we combine the frames,
        so we are facing a higher volume in the
        segments in which we have one single
        audio and different volumes when more
        than one are being played together...

        Thats why I have this code commented by
        now.
        """
        if False:
            # We attenuate the audios
            # attenuation_gain = 1.0 / math.sqrt(len(arrays))
            # arrays = [
            #     arr * attenuation_gain
            #     for arr in arrays
            # ]

            # Sum all the sounds
            mix = np.sum(stacked, axis = 0)

            # if do_normalize:
            #     # Avoid distortion and saturation
            #     mix /= len(stacked)

            # # Avoid clipping
            # mix = np.clip(mix, -1.0, 1.0)
        else:
            mix = np.sum(stacked, axis = 0)

        out = AudioFrame.from_ndarray(
            array = mix,
            format = format,
            layout = layout
        )
        out.sample_rate = sample_rate

        return out


"""
The track module, that includes the
different types of tracks we accept for
the editor timeline.

A track is a sub-timeline in which we
can place the media that is accepted by
that track. The empty areas will be
fulfilled with imaginary 'empty' medias.
"""
# from yta_editor.track.audio import AudioTrack
# from yta_editor.track.video import VideoTrack


# __all__ = [
#     'AudioTrack',
#     'VideoTrack'
# ]


from yta_editor.utils.silence import generate_silent_frames
from yta_editor.utils.frame_wrapper import AudioFrameWrapped, VideoFrameWrapped
from yta_editor.utils.frame_generator import VideoFrameGenerator
from yta_video_frame_time.t_fraction import fps_to_time_base
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union
from abc import ABC


NON_LIMITED_EMPTY_PART_END = 999
"""
A value to indicate that the empty part
has no end because it is in the last
position and there is no video after it.
"""

class _Part(ABC):
    """
    Abstract class to represent an element
    that is on the track, that can be an
    empty space or a vide or audio. This
    class must be inherited by our own
    custom part classes.
    """

    @property
    def is_empty_part(
        self
    ) -> bool:
        """
        Flag to indicate if the part is an empty part,
        which means that there is no media associated
        but an empty space.
        """
        return self.media is None

    def __init__(
        self,
        track: Union['AudioTrack', 'VideoTrack'],
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction],
        media: Union['AudioOnTrack', 'VideoOnTrack', None] = None
    ):
        ParameterValidator.validate_mandatory_positive_number('start', start, do_include_zero = True)
        ParameterValidator.validate_mandatory_positive_number('end', end, do_include_zero = False)
        ParameterValidator.validate_instance_of('media', media, ['AudioOnTrack', 'VideoOnTrack'])

        self._track: Union['AudioTrack', 'VideoTrack'] = track
        """
        The instance of the track this part belongs
        to.
        """
        self.start: Fraction = Fraction(start)
        """
        The start 't' time moment of the part.
        """
        self.end: Fraction = Fraction(end)
        """
        The end 't' time moment of the part.
        """
        self.media: Union['AudioOnTrack', 'VideoOnTrack', None] = media
        """
        The media associated, if existing, or
        None if it is an empty space that we need
        to fulfill.
        """

class _PartWithAudio(_Part):
    """
    TODO: Explain
    """

    @property
    def _silent_frames(
        self
    ) -> list[AudioFrameWrapped]:
        """
        *For internal use only*

        The silent frames we need to fulfill the
        space when an empty part. Property to
        avoid creating these silent frames again
        and again.
        """
        if not hasattr(self, '__silent_frames'):
            self.__silent_frames = generate_silent_frames(
                video_fps = self._track.fps,
                audio_fps = self._track.audio_fps,
                audio_samples_per_frame = self._track.audio_samples_per_frame,
                # TODO: Where do this 2 formats come from (?)
                layout = self._track.audio_layout,
                format = self._track.audio_format
            )

        return self.__silent_frames

    def __init__(
        self,
        track: Union['AudioTrack', 'VideoTrack'],
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction],
        media: Union['AudioOnTrack', 'VideoOnTrack', None] = None
    ):
        ParameterValidator.validate_instance_of('media', media, ['AudioOnTrack', 'VideoOnTrack'])

        super().__init__(
            track = track,
            start = start,
            end = end,
            media = media
        )

    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps = Union[int, float, Fraction]
    ):
        """
        Iterate over all the audio frames that
        exist at the time moment 't' provided.
        """
        frames = []
        if not self.is_empty_part:
            # TODO: What do we do in this case (?)
            frames = list(self.media.get_audio_frames_at(t, video_fps))

            if len(frames) == 0:
                print(f'   [ERROR] Audio frame {str(float(t))} was not obtained')
            else:
                frames = [
                    AudioFrameWrapped(
                        frame = frame,
                        is_from_empty_part = False
                    )
                    for frame in frames
                ]

        # This could be because is empty part or
        # because we couldn't obtain the frames
        frames = (
            self._silent_frames
            if len(frames) == 0 else
            frames
        )

        for frame in frames:
            yield frame

class _PartWithVideo(_Part):
    """
    TODO: Explain
    """

    @property
    def _frame_not_found(
        self
    ):
        """
        *For internal use only*

        The frame we should send when we didn't
        obtain a frame from a source that should
        be available. Property to avoid creating
        this frame again and again.

        TODO: This should not happen, it is an
        internal way to point an error.
        """
        if not hasattr(self, '__empty_frame'):
            self.__frame_not_found = VideoFrameGenerator.background.full_red(
                size = self._track.size,
                time_base = fps_to_time_base(self._track.fps)
            )

        return self.__frame_not_found

    @property
    def _empty_frame(
        self
    ):
        """
        *For internal use only*

        The frame we should send when an empty
        part is playing. Property to avoid
        creating this frame again and again.
        """
        if not hasattr(self, '__empty_frame'):
            self.__empty_frame = VideoFrameGenerator.background.full_black(
                size = self._track.size,
                time_base = fps_to_time_base(self._track.fps),
                transparency = 1.0
            )

        return self.__empty_frame

    def __init__(
        self,
        track: 'VideoTrack',
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction],
        media: Union['VideoOnTrack', None] = None
    ):
        ParameterValidator.validate_instance_of('media', media, 'VideoOnTrack')

        super().__init__(
            track = track,
            start = start,
            end = end,
            media = media
        )

    def get_video_frame_at_t(
        self,
        t: Union[int, float, Fraction]
    ) -> 'VideoFrameWrapped':
        """
        Get the frame that must be displayed at 
        the given 't' time moment.
        """
        is_from_empty_part = False
        if self.is_empty_part == True:
            frame = self._empty_frame

            is_from_empty_part = True
        else:
            # TODO: This can be None, why? I don't know...
            frame = self.media.get_video_frame_at(t)

            if frame is None:
                print(f'   [ERROR] Frame {str(float(t))} was not obtained')

            frame = (
                # I'm using a red full frame to be able to detect
                # fast the frames that were not available, but
                # I need to find the error and find a real solution
                # TODO: This shouldn't happen, its an error
                self._frame_not_found
                if frame is None else
                frame
            )

        # TODO: What about the 'format' (?)
        # TODO: Maybe I shouldn't set the 'time_base'
        # here and do it just in the Timeline 'render'
        #return get_black_background_video_frame(self._track.size)
        # TODO: This 'time_base' maybe has to be related
        # to a Timeline general 'time_base' and not the fps
        frame = VideoFrameWrapped(
            frame = frame,
            is_from_empty_part = is_from_empty_part
        )

        # TODO: This should not happen because of
        # the way we handle the videos here but the
        # video could send us a None frame here, so
        # do we raise exception (?)
        if frame._frame is None:
            #frame = get_black_background_video_frame(self._track.size)
            # TODO: By now I'm raising exception to check if
            # this happens or not because I think it would
            # be malfunctioning
            raise Exception(f'Video is returning None video frame at t={str(t)}.')
        
        """
        The 'track' (because of the 'timeline' it belongs
        to) has a size, so we must respect, so we need to
        use a strategy to make it fit the desired size.
        """
        # TODO: Maybe this sould be done in the 'timeline'
        # that combines the frames and not here...
        return frame


"""
Some examples of valid video formats in
ffmpeg library:
- YUV formats: yuv420p, yuv422p, yuv444p, yuva420p, etc.
- RGB formats: rgb24, rgba, bgra, bgr0, etc.
- Grayscale formats: gray, ya8, etc.
- Special or compressed formats: nv12, nv21, p010le, etc.
- 10/12 bits formats: yuv420p10le, yuv444p12le, etc.
"""
from av.video.format import VideoFormat


# TODO: Maybe 'rgb24' should be removed in
# favor of yuv420p which is more compatible
ACCEPTED_OUTPUT_FORMATS = ['rgb24', 'rgba', 'yuv420p', 'yuva420p']
"""
The only pyav and ffmpeg video formats we
accept to export a video.
"""
VALID_OUTPUTS = {
    # Non-alpha below
    'rgb24': {'mp4', 'mov', 'avi', 'mkv', 'webm'},
    'yuv420p': {'mp4', 'mov', 'avi', 'mkv', 'webm'},
    # Alpha below
    'rgba': {'mov', 'avi', 'mkv'},
    'yuva420p': {'mov', 'webm', 'mkv'}
}
"""
A dictionary containing the video file
extensions that can handle the different
video formats we accept.
"""


class Validator:
    """
    Class to simplify validation with static
    methods.
    """

    @staticmethod
    def validate_output_video_format(
        format: str
    ) -> None:
        """
        Raise an exception if the provided
        'format' is not one of the formats
        we accept as output:
        - `rgb24`
        - `rgba`
        - `yuv420p`
        - `yuva420p`
        """
        if not format.lower() in ACCEPTED_OUTPUT_FORMATS:
            raise Exception(f'The provided format "{format}" is not one of our accepted output video formats: {", ".join(ACCEPTED_OUTPUT_FORMATS)}')
        
    @staticmethod
    def validate_filename_for_output_video_format(
        filename: str,
        format: str
    ) -> None:
        """
        Raise an exception if the extension of
        the provided 'filename' is not valid for
        the also given 'format', or if that
        'format' is not even valid by itself.
        """
        Validator.validate_output_video_format(format)

        valid_formats = VALID_OUTPUTS.get(format, set())
        if filename.split('.')[-1] not in valid_formats:
            raise Exception(f'The extension of the provided filename "{filename}" is not valid for the also given format "{format}". The filename must have one of these extensions: .{", .".join(valid_formats)}')

    
def validate_ffmpeg_video_format(
    format: str
) -> None:
    """
    Raise an exception if the provided
    'format' is not a valid ffmpeg video
    format.

    The video format will be validated
    according to the version of the ffmpeg
    library that is installed in this
    system.
    """
    if not format.lower() in VideoFormat.formats.values():
        raise Exception(f'The provided format "{format}" is not a valid video frame format for the current ffmpeg library (if installed).')


"""
If we have a video placed in a timeline,
starting at the t=2s and the video lasts
2 seconds, the `t` time range in which the
video is playing is `[2s, 4s]`, so here 
you have some examples with global `t` 
values:
- `t=1`, the video is not playing because
it starts at `t=2`
- `t=3`, the video is playing, it started
at `t=2` and it has been playing during 1s
- `t=5`, the video is not playing because
it started at `t=2`, lasting 2s, so it
finished at `t=4`
"""
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from quicktions import Fraction
from typing import Union
from abc import ABC


# TODO: Maybe rename to something like 
# 'MediaWithStartAndEnd' because it is not
# only for the objects that we place on a
# track but also for other conditions in 
# which we need the 'start' and 'end'
# paramters (like when joined with a
# transition)
class _MediaOnTrack(ABC):
    """
    Class to be inherited by any media class
    that will be placed on a track and should
    manage this condition.
    """

    @property
    def end(
        self
    ) -> Fraction:
        """
        The end time moment 't' of the audio once
        once its been placed on the track, which
        is affected by the audio duration and its
        start time moment on the track.

        This end is different from the audio end.

        The formula:
        - `self.start + self.duration`
        """
        return self.start + self.duration
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        Shortcut to the duration of the media we
        are applying on the track.

        The formula:
        - `self.media.duration`
        """
        return self.media.duration

    def __init__(
        self,
        media: Union['AudioFileMedia', 'VideoFileMedia', 'VideoColorMedia', 'VideoImageMedia'],
        start: Union[int, float, Fraction] = 0.0
    ):
        ParameterValidator.validate_mandatory_instance_of('media', media, ['AudioFileMedia', 'VideoFileMedia', 'VideoColorMedia', 'VideoImageMedia'])
        ParameterValidator.validate_mandatory_positive_number('start', start, do_include_zero = True)

        self.media: Union['AudioFileMedia', 'VideoFileMedia', 'VideoColorMedia', 'VideoImageMedia'] = media
        """
        The media source, with all its properties,
        that is placed in the timeline.
        """
        self.start: Fraction = Fraction(start)
        """
        The time moment in which the media should
        start playing, within the timeline.

        This is the time respect to the timeline
        and its different from the media `start`
        time, which is related to the file.
        """
    
    def _get_t(
        self,
        t: Union[int, float, Fraction]
    ) -> float:
        """
        The media 't' time moment for the given
        global 't' time moment. This 't' is the one
        to use inside the media content to display
        its frame.
        """
        # TODO: Should we make sure 't' is truncated (?)
        return t - self.start

    def is_playing(
        self,
        t: Union[int, float, Fraction]
    ) -> bool:
        """
        Check if this media is playing at the general
        't' time moment, which is a global time moment
        for the whole project.
        """
        # TODO: Should we make sure 't' is truncated (?)
        return self.start <= t < self.end
    
class _MediaOnTrackWithAudio(_MediaOnTrack):
    """
    Class that implements the ability of
    getting audio frames. This class must
    be inherited by any other class that
    has this same ability.
    """

    def __init__(
        self,
        media: Union['AudioFileMedia', 'VideoFileMedia', 'VideoColorMedia', 'VideoImageMedia'],
        start: Union[int, float, Fraction] = 0.0
    ):
        super().__init__(
            media = media,
            start = start
        )

    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction]
    ):
        """
        Get the audio frames that must be played at
        the 't' time moment provided, that could be
        None if the audio is not playing at that
        moment.

        This method will return an empty array if
        no audio frames found in that 't' time
        moment, or an iterator if yes.

        The `t` time moment provided must be a global
        time value pointing to a moment in the general
        timeline. That value will be transformed to the
        internal `t` value to get the frame.
        """
        # TODO: Use 'T' here to be precise or the
        # argument itself must be precise (?)
        frames = (
            self.media.get_audio_frames_at_t(
                t = self._get_t(t),
                video_fps = video_fps
            )
            if self.is_playing(t) else
            []
        )

        for frame in frames:
            yield frame

class _MediaOnTrackWithVideo(_MediaOnTrack):
    """
    Class that implements the ability of
    getting video frames. This class must
    be inherited by any other class that
    has this same ability.
    """

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        Shortcut to the size of the media.
        """
        return self.media.size

    def __init__(
        self,
        media: Union['VideoFileMedia', 'VideoColorMedia', 'VideoImageMedia'],
        start: Union[int, float, Fraction] = 0.0
    ):
        super().__init__(
            media = media,
            start = start
        )

    def get_video_frame_at(
        self,
        t: Union[int, float, Fraction]
    ) -> Union[VideoFrame, None]:
        """
        Get the frame for the 't' time moment provided,
        that could be None if the video is not playing
        in that moment.

        The `t` time moment provided must be a global
        time value pointing to a moment in the general
        timeline. That value will be transformed to the
        internal `t` value to get the frame.
        """
        # TODO: Use 'T' here to be precise or the
        # argument itself must be precise (?)
        return (
            self.media.get_video_frame_at_t(
                t = self._get_t(t)
            )
            if self.is_playing(t) else
            None
        )


"""
This module must be @deprecated due to the new
version we have that handles it with OpenGL.
This one is using numpy and CPU.
"""
from yta_editor.track.media.video import VideoOnTrack
from yta_editor.media.video import _VideoMedia
from yta_editor.media.transition.classes import AlphaBlendTransition
from yta_video_frame_time.t_fraction import get_ts, fps_to_time_base, T
from yta_video_pyav.writer import VideoWriter
from yta_video_pyav.settings import Settings
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from quicktions import Fraction
from typing import Union

import numpy as np


class TransitionMedia:
    """
    Class to wrap a transition clip, that is a
    clip in which 2 clips are joined with another
    transition clip.

    TODO: This below is temporary
    The clips will be placed consecutive and the
    transition will be played in between, joining
    both.
    """

    @property
    def _start(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The moment in which the transition clip starts
        being played, that will be in some moment of
        the first clip, and it should be also the 
        moment in which the second clip starts being
        played (with the corresponding transition
        blending).
        """
        # TODO: This value is only valid for the transition
        # applied in the middle with 50%-50% strategy
        return self._clip_a.duration - self.duration
    
    @property
    def _end(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The moment in which the whole clip built with
        the transition in the middle must stop being
        played, so the second clip has finished.
        """
        return self._clip_b.end

    def __init__(
        self,
        # TODO: This '_MediaOnTrack' has to be renamed 
        # because here it is not on a track but the class
        # we need is this one
        # TODO: Maybe it has to be just a Media, not OnTrack
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        # TODO: Apply Union[AlphaBlendTransition, ...]
        transition: AlphaBlendTransition,
        duration: float,
    ):
        ParameterValidator.validate_subclass_of('clip_a', clip_a, _VideoMedia)
        ParameterValidator.validate_subclass_of('clip_b', clip_b, _VideoMedia)

        if (
            clip_a.duration < duration or
            clip_b.duration < duration
        ):
            raise Exception('The duration of one of the clips is smaller than the transition duration requested.')

        # TODO: Apply Union[AlphaBlendTransition, ...]
        self._transition: AlphaBlendTransition = transition
        """
        The configuration we want to apply in the transition.
        """
        self.duration: float = duration
        """
        The duration of the transition.
        """
        # TODO: Rename different than OnTrack
        self._clip_a: VideoOnTrack = VideoOnTrack(
            media = clip_a,
            start = 0
        )
        """
        The first clip to join in with the transition.
        """
        # TODO: Rename different than OnTrack
        self._clip_b: VideoOnTrack = VideoOnTrack(
            media = clip_b,
            start = self._start
        )
        """
        The second clip to join in with the transition.
        """
    
    # TODO: We manually handle the time in which
    # we are applying the transition and use the
    # blend factor only there, so this method is
    # unnecessary I think unless we use a curve
    # and not a linear blend factor
    def get_blend_factor_at_t(
        self,
        t: float
    ) -> float:
        """
        Get the weight of the second clip at the
        't' time moment provided as a value in the
        [0.0, 1.0] range.

        The `t` time moment provided must be a 
        time moment in between 0 and the end of
        the whole transition.

        TODO: What (?)
        The `t` is the time moment since the start
        of the transition.
        """
        if t < 0:
            return 0.0
        
        if t > self.duration:
            return 1.0
        
        """
        By now, the transition we are applying is
        as simple as being in the middle of the 2
        clips provided and lasting the 'duration'
        provided, being executed with a linear t.
        """

        # TODO: The ratios are to adjust the internal
        # duration but we are not using them by now
        # Normalización según start/end ratios
        # ratio_range = self.end_ratio - self.start_ratio
        # norm_t = (t / self.duration - self.start_ratio) / ratio_range
        # norm_t = max(0.0, min(1.0, norm_t))

        norm_t = (t / self.duration)
        norm_t = max(0.0, min(1.0, norm_t))

        # The curve is a param to adjust the transition
        # speed that we are not handling by now
        curve = 0.0
        # TODO: By now we have one simple way of
        # handling the 't' value
        norm_t = norm_t ** (1 + curve)

        # # Easing
        # if self.ease == TransitionEase.EASE_IN:
        #     norm_t = norm_t ** (1 + self.curve)
        # elif self.ease == TransitionEase.EASE_OUT:
        #     norm_t = 1 - (1 - norm_t) ** (1 + self.curve)
        # elif self.ease == TransitionEase.EASE_IN_OUT:
        #     # curva suave simétrica
        #     norm_t = 0.5 * (1 - (1 - 2 * norm_t) ** (1 + self.curve)) if norm_t < 0.5 else \
        #              0.5 + 0.5 * (2 * norm_t - 1) ** (1 + self.curve)

        return norm_t
    
    def get_video_frame_at(
        self,
        t: Union[int, float, Fraction],
        do_apply_filters: bool = False
    ) -> VideoFrame:
        """
        The `t` time moment provided must be a global
        time value pointing to a moment in the general
        timeline. That value will be transformed to the
        internal `t` value to get the frame.
        """
        frame_a = self._clip_a.get_video_frame_at(t)
        frame_b = self._clip_b.get_video_frame_at(t)

        if (
            frame_a is None and
            frame_b is None
        ):
            print('Both None, wtf')
            #return None

        """
        There is a moment in which we are not in
        the transition section but on the clips
        so we just need to return the clip frames.
        """
        if frame_b is None:
            # Not transition time, just one clip
            out_frame = frame_a
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame
        
        if frame_a is None:
            # Not transition time, just one clip
            out_frame = frame_b
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame

        # TODO: By now I'm forcing a simple crossfade
        # transition just to validate it works
        # TODO: This has to be more complex and also
        # using curves
        t_progress = (t - self._start) / self.duration
        print(f' ------->  T progress: {str(t_progress)}')

        # TODO: What about 'get_blend_factor_at_t' (?)

        result_frame = VideoFrame.from_ndarray(
            array = (
                self._transition.blend_frames(
                    frame_a = frame_a.to_ndarray(format = 'rgb24').astype(np.float32),
                    frame_b = frame_b.to_ndarray(format = 'rgb24').astype(np.float32),
                    t_progress = t_progress
                )
            ).astype(np.uint8),
            format = 'rgb24'
        )
        # The 'pts' and that is not important here but...
        result_frame.pts = None
        result_frame.time_base = Fraction(1, 60)

        return result_frame
    
    # def get_audio_frames_at(
    #     self,
    #     t: Union[int, float, Fraction],
    #     video_fps: Union[int, float, Fraction, None] = None,
    #     do_apply_filters: bool = True
    # ):
    #     """
    #     Get the sequence of audio frames for the 
    #     given video 't' time moment, using the
    #     audio cache system.

    #     This is useful when we want to write a
    #     video frame with its audio, so we obtain
    #     all the audio frames associated to it
    #     (remember that a video frame is associated
    #     with more than 1 audio frame).
    #     """
    #     frame_a = self._clip_a.get_audio_frames_at(t)
    #     frame_b = self._clip_b.get_video_frame_at(t)

    #     if (
    #         frame_a is None and
    #         frame_b is None
    #     ):
    #         print('Both None, wtf')
    #         #return None

    #     """
    #     There is a moment in which we are not in
    #     the transition section but on the clips
    #     so we just need to return the clip frames.
    #     """
    #     if frame_b is None:
    #         # Not transition time, just one clip
    #         out_frame = frame_a
    #         out_frame.pts = None
    #         out_frame.time_base = Fraction(1, 60)

    #         return out_frame
        
    #     if frame_a is None:
    #         # Not transition time, just one clip
    #         out_frame = frame_b
    #         out_frame.pts = None
    #         out_frame.time_base = Fraction(1, 60)

    #         return out_frame

    #     # TODO: By now I'm forcing a simple crossfade
    #     # transition just to validate it works
    #     # TODO: This has to be more complex and also
    #     # using curves
    #     t_progress = (t - self._start) / self.duration
    #     print(f' ------->  T progress: {str(t_progress)}')

    #     # TODO: What about 'get_blend_factor_at_t' (?)

    #     # TODO: How to mix the audio frames (?)
    #     raise Exception('Not ready yet')
    #     # numpy_frame_a = frame_a.to_ndarray(format = 'rgb24').astype(np.float32)
    #     # numpy_frame_b = frame_b.to_ndarray(format = 'rgb24').astype(np.float32)

    #     # result_frame = VideoFrame.from_ndarray(
    #     #     array = (
    #     #         # Replace by specific alpha blending strategy
    #     #         (1 - t_progress) * numpy_frame_a + t_progress * numpy_frame_b
    #     #     ).astype(np.uint8),
    #     #     format = 'rgb24'
    #     # )
    #     # # The 'pts' and that is not important here but...
    #     # result_frame.pts = None
    #     # result_frame.time_base = Fraction(1, 60)

    #     # return result_frame
    
    def save_as(
        self,
        output_filename: str,
        video_size: tuple[int, int] = None,
        video_fps: float = None,
        video_codec: str = None,
        video_pixel_format: str = None,
        audio_codec: str = None,
        audio_sample_rate: int = None,
        audio_layout: str = None,
        audio_format: str = None,
        do_apply_video_filters: bool = True,
        do_apply_audio_filters: bool = True
    ) -> str:
        """
        Save the file as 'output_filename'.

        This method is useful if you want to apply
        some filter and then save the video with
        those filters applied into a new one, maybe
        with a new pixel format and/or code. You can
        prepare alpha transitions, etc.
        """
        video_size = (
            getattr(self, 'size', Settings.DEFAULT_VIDEO_SIZE.value)
            if video_size is None else
            video_size
        )

        video_fps = (
            getattr(self, 'fps', Settings.DEFAULT_VIDEO_FPS.value)
            if video_fps is None else
            video_fps
        )

        video_codec = (
            getattr(self, 'codec_name', Settings.DEFAULT_VIDEO_CODEC.value)
            if video_codec is None else
            video_codec
        )

        video_pixel_format = (
            getattr(self, 'pixel_format', Settings.DEFAULT_PIXEL_FORMAT.value)
            if video_pixel_format is None else
            video_pixel_format
        )

        audio_codec = (
            getattr(self, 'audio_codec_name', Settings.DEFAULT_AUDIO_CODEC.value)
            if audio_codec is None else
            audio_codec
        )

        audio_sample_rate = (
            getattr(self, 'audio_fps', Settings.DEFAULT_AUDIO_FPS.value)
            if audio_sample_rate is None else
            audio_sample_rate
        )

        audio_layout = (
            getattr(self, 'audio_layout', Settings.DEFAULT_AUDIO_LAYOUT.value)
            if audio_layout is None else
            audio_layout
        )

        audio_format = (
            getattr(self, 'audio_format', Settings.DEFAULT_AUDIO_FORMAT.value)
            if audio_format is None else
            audio_format
        )

        writer = VideoWriter(output_filename)

        # TODO: This has to be dynamic according to the
        # video we are writing (?)
        writer.set_video_stream(
            codec_name = video_codec,
            fps = video_fps,
            size = video_size,
            pixel_format = video_pixel_format
        )
        
        writer.set_audio_stream(
            codec_name = audio_codec,
            fps = audio_sample_rate,
            layout = audio_layout,
            format = audio_format
        )

        # TODO: Maybe we need to reformat or something
        # if some of the values changed, such as fps,
        # audio sample rate, etc. (?)

        time_base = fps_to_time_base(video_fps)
        audio_time_base = fps_to_time_base(audio_sample_rate)

        for t in get_ts(0, self._end, video_fps):
            frame = self.get_video_frame_at(
                t = t,
                do_apply_filters = do_apply_video_filters
            )

            # TODO: What if 'frame' is None (?)
            if frame is None:
                print(f'   [ERROR] Frame not found at t:{float(t)}')
                continue

            writer.mux_video_frame(
                frame = frame
            )

            frame.time_base = time_base
            frame.pts = T(t, time_base).truncated_pts

            # TODO: Make this work
            # audio_pts = 0
            # for audio_frame in self.get_audio_frames_at_t(
            #     t = t,
            #     video_fps = video_fps,
            #     do_apply_filters = do_apply_audio_filters
            # ):
            #     # TODO: 'audio_frame' could be None or []
            #     # here if no audio channel
            #     if audio_frame is None:
            #         # TODO: Generate silence audio to cover the
            #         # whole video frame (?)
            #         pass
                
            #     # We need to adjust our output elements to be
            #     # consecutive and with the right values
            #     # TODO: We are using int() for fps but its float...
            #     audio_frame.time_base = audio_time_base
            #     audio_frame.pts = audio_pts

            #     # We increment for the next iteration
            #     audio_pts += audio_frame.samples

            #     writer.mux_audio_frame(audio_frame)

        writer.mux_video_frame(None)
        writer.mux_audio_frame(None)
        writer.output.close()

        return output_filename

"""
Davinci Resolve has these 3 types of transition
strategies:
- Start on Cut → La transición empieza justo en el corte, extendiéndose hacia adelante.
- Center on Cut → La transición se centra en el corte (la mitad sobre el primer clip y la mitad sobre el segundo).
- End on Cut → La transición termina justo en el corte, extendiéndose hacia atrás.
"""


from yta_editor.utils import VideoUtils
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from typing import Union

import numpy as np


IS_FROM_EMPTY_PART_METADATA = 'is_from_empty_part'
"""
Metadata key to indicate if the frame
has been generated by an empty part
and should be ignored when trying to
combine with others.
"""

class _FrameWrappedBase:
    """
    Class to wrap video and audio frames from
    the pyav library but to support a metadata
    field to inject some information we need
    when processing and combining them.
    """

    @property
    def frame(
        self
    ) -> VideoFrame:
        """
        Get the original pyav VideoFrame.
        """
        return self._frame

    @property
    def is_from_empty_part(
        self
    ) -> bool:
        """
        Flag to indicate if the frame comes from
        an empty part or not, that will be done
        by checking the 'is_from_empty_part' 
        attribute in the metadata.

        This will return True only if the value
        is set and it is set as 'True'.
        """
        is_from_empty_part = self.metadata.get(IS_FROM_EMPTY_PART_METADATA, None)

        return (
            is_from_empty_part == 'True'
            if is_from_empty_part is not None else
            is_from_empty_part
        )

    def __init__(
        self,
        frame: Union[VideoFrame, AudioFrame],
        metadata: dict = {}
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, [VideoFrame, AudioFrame])
        ParameterValidator.validate_mandatory_dict('metadata', metadata)

        self._frame: Union[VideoFrame, AudioFrame] = frame
        """
        The VideoFrame or AudioFrame pyav instance.
        """
        self.metadata: dict = metadata or {}
        """
        The metadata we want to include with the
        frame.
        """

    def __getattr__(
        self,
        name
    ):
        return getattr(self._frame, name)

    def __setattr__(
        self,
        name,
        value
    ):
        super().__setattr__(name, value)
        #setattr(self._frame, name, value)
        # if name in ('_frame', 'metadata'):
        #     super().__setattr__(name, value)
        # else:
        #     setattr(self._frame, name, value)

    def __repr__(
        self
    ):
        cname = self.__class__.__name__
        return f'<{cname} metadata={self.metadata} frame={self._frame!r}>'
    
    def set_is_from_empty_part(
        self,
        is_from_empty_part: bool = True
    ) -> None:
        """
        Add the metadata information to indicate
        that this is a frame that comes from an 
        empty part.
        """
        self.metadata[IS_FROM_EMPTY_PART_METADATA] = (
            'True'
            if is_from_empty_part else
            'False'
        )
    
class VideoFrameWrapped(_FrameWrappedBase):
    """
    Class to wrap video frames from the pyav
    library but to support a metadata field
    to inject some information we need when
    processing and combining them.
    """

    @property
    def has_alpha_layer(
        self
    ) -> bool:
        """
        Flag to indicate if the frame has an alpha
        layer or not, that is defined by an 'a' in
        its format name.
        """
        return VideoUtils.video.alpha.videoframe_has_alpha_layer(self._frame)
    
    @property
    def has_alpha_pixels(
        self
    ) -> bool:
        """
        Flag to indicate if the frame comes with
        at least one pixel that has some kind of
        transparency and has to be considered for
        the alpha blending process because of that.
        """
        return (
            self.has_alpha_layer and
            VideoUtils.video.alpha.numpy_videoframe_has_transparent_pixels(self.as_rgba_numpy)
        )
    
    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the frame expressed as (width,
        height).
        """
        return (self.frame.width, self.frame.height)

    @property
    def as_rgb_numpy(
        self
    ) -> 'np.ndarray':
        """
        Get the video frame but as a numpy array with
        the 'rgb24' format, that doesn't include the
        alpha channel.

        The 'rgb24' format is perfect for video
        edition including not the alpha channel.
        """
        return self.frame.to_ndarray(format = 'rgb24')
    
    @property
    def as_rgba_numpy(
        self
    ) -> 'np.ndarray':
        """
        Get the video frame but as a numpy array with
        the 'rgba' format, that includes the alpha
        channel.

        The 'rgba' format is perfect for video edition
        including the alpha channel.
        """
        return self.frame.to_ndarray(format = 'rgba')

    def __init__(
        self,
        frame: VideoFrame,
        metadata: dict = {},
        is_from_empty_part: Union[bool, None] = None,
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)

        super().__init__(
            frame = frame,
            metadata = metadata
        )

        if is_from_empty_part is not None:
            self.set_is_from_empty_part(is_from_empty_part)
        
# TODO: Maybe move? But careful with
# cyclic import issues
class AudioFrameWrapped(_FrameWrappedBase):
    """
    Class to wrap audio frames from the pyav
    library but to support a metadata field
    to inject some information we need when
    processing and combining them.
    """

    def __init__(
        self,
        frame: AudioFrame,
        metadata: dict = {},
        is_from_empty_part: bool = False
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, AudioFrame)

        super().__init__(
            frame = frame,
            metadata = metadata
        )

        if is_from_empty_part:
            self.set_is_from_empty_part()


from yta_editor.media.abstract import _Media
from yta_editor.sources.abstract import _VideoSource
from yta_editor.sources.video import VideoFileSource, VideoImageSource, VideoColorSource, VideoNumpySource
from yta_editor.decorators import with_t_adjusted_to_media
from yta_editor.utils.effects import apply_audio_effects_to_frame_at_t, apply_video_effects_to_frame_at_t
from yta_video_frame_time.t_fraction import get_ts, fps_to_time_base, T
from yta_video_pyav.writer import VideoWriter
from yta_video_pyav.settings import Settings
from yta_video_pyav.reader.filter.dataclass import GraphFilter, GraphFilters
from yta_video_opengl.nodes import TimedNode
from yta_colors import Color
from av.video.frame import VideoFrame
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union

    
class _VideoMedia(_Media):

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the original source, expressed
        as (width, height). Check the 'output_size'
        to know the size requested by the user to
        be the size of each video frame.
        """
        return self.source.size
    
    def __init__(
        self,
        source: _VideoSource,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None
    ):
        super().__init__(
            source = source,
            start = start,
            end = end
        )

    def add_effect(
        self,
        effect: TimedNode
    ) -> 'Video':
        """
        Add the provided 'effect' to the video.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, 'TimedNode')

        self._effects.add_effect(effect)
        
        return self
    
    @with_t_adjusted_to_media
    def get_video_frame_at_t(
        self,
        t: Union[int, float, Fraction],
        do_apply_filters: bool = True
    ) -> Union[VideoFrame, None]:
        """
        Get the video frame with the given 't' time
        moment, using the video cache system.
        """
        print(f'Getting frame from {str(float(t + self.start))} that is actually {str(float(t))}')
        
        frame = self.source.get_video_frame_at_t(
            t = t,
            size = self.output_size,
            do_apply_filters = do_apply_filters
        )

        return (
            apply_video_effects_to_frame_at_t(
                effects_stack = self._effects,
                frame = frame,
                t = t
            )
            if frame is not None else
            None
        )
    
    @with_t_adjusted_to_media
    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction, None] = None,
        do_apply_filters: bool = True
    ):
        """
        Get the sequence of audio frames for the 
        given video 't' time moment, using the
        audio cache system.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        video_fps = (
            self.fps
            if video_fps is None else
            video_fps
        )

        print(f'Getting audio frames from {str(float(t + self.start))} that is actually {str(float(t))}')
        for frame in self.source.get_audio_frames_at_t(
            t = t,
            video_fps = video_fps,
            do_apply_filters = do_apply_filters
        ):
            yield apply_audio_effects_to_frame_at_t(
                effects_stack = self._effects,
                frame = frame,
                t = t
            )

    def save_as(
        self,
        output_filename: str,
        video_size: tuple[int, int] = None,
        video_fps: float = None,
        video_codec: str = None,
        video_pixel_format: str = None,
        audio_codec: str = None,
        audio_sample_rate: int = None,
        audio_layout: str = None,
        audio_format: str = None,
        do_apply_video_filters: bool = True,
        do_apply_audio_filters: bool = True
    ) -> str:
        """
        Save the file as 'output_filename'.

        This method is useful if you want to apply
        some filter and then save the video with
        those filters applied into a new one, maybe
        with a new pixel format and/or code. You can
        prepare alpha transitions, etc.
        """
        video_size = (
            getattr(self, 'size', Settings.DEFAULT_VIDEO_SIZE.value)
            if video_size is None else
            video_size
        )

        video_fps = (
            getattr(self, 'fps', Settings.DEFAULT_VIDEO_FPS.value)
            if video_fps is None else
            video_fps
        )

        video_codec = (
            getattr(self, 'codec_name', Settings.DEFAULT_VIDEO_CODEC.value)
            if video_codec is None else
            video_codec
        )

        video_pixel_format = (
            getattr(self, 'pixel_format', Settings.DEFAULT_PIXEL_FORMAT.value)
            if video_pixel_format is None else
            video_pixel_format
        )

        audio_codec = (
            getattr(self, 'audio_codec_name', Settings.DEFAULT_AUDIO_CODEC.value)
            if audio_codec is None else
            audio_codec
        )

        audio_sample_rate = (
            getattr(self, 'audio_fps', Settings.DEFAULT_AUDIO_FPS.value)
            if audio_sample_rate is None else
            audio_sample_rate
        )

        audio_layout = (
            getattr(self, 'audio_layout', Settings.DEFAULT_AUDIO_LAYOUT.value)
            if audio_layout is None else
            audio_layout
        )

        audio_format = (
            getattr(self, 'audio_format', Settings.DEFAULT_AUDIO_FORMAT.value)
            if audio_format is None else
            audio_format
        )

        writer = VideoWriter(output_filename)

        # TODO: This has to be dynamic according to the
        # video we are writing (?)
        writer.set_video_stream(
            codec_name = video_codec,
            fps = video_fps,
            size = video_size,
            pixel_format = video_pixel_format
        )
        
        writer.set_audio_stream(
            codec_name = audio_codec,
            fps = audio_sample_rate,
            layout = audio_layout,
            format = audio_format
        )

        # TODO: Maybe we need to reformat or something
        # if some of the values changed, such as fps,
        # audio sample rate, etc. (?)

        time_base = fps_to_time_base(video_fps)
        audio_time_base = fps_to_time_base(audio_sample_rate)

        for t in get_ts(0, self.end, video_fps):
            frame = self.get_video_frame_at_t(
                t = t,
                do_apply_filters = do_apply_video_filters
            )

            # TODO: What if 'frame' is None (?)
            if frame is None:
                print(f'   [ERROR] Frame not found at t:{float(t)}')
                continue

            writer.mux_video_frame(
                frame = frame
            )

            frame.time_base = time_base
            frame.pts = T(t, time_base).truncated_pts

            audio_pts = 0
            for audio_frame in self.get_audio_frames_at_t(
                t = t,
                video_fps = video_fps,
                do_apply_filters = do_apply_audio_filters
            ):
                # TODO: 'audio_frame' could be None or []
                # here if no audio channel
                if audio_frame is None:
                    # TODO: Generate silence audio to cover the
                    # whole video frame (?)
                    pass
                
                # We need to adjust our output elements to be
                # consecutive and with the right values
                # TODO: We are using int() for fps but its float...
                audio_frame.time_base = audio_time_base
                audio_frame.pts = audio_pts

                # We increment for the next iteration
                audio_pts += audio_frame.samples

                writer.mux_audio_frame(audio_frame)

        writer.mux_video_frame(None)
        writer.mux_audio_frame(None)
        writer.output.close()

        return output_filename
    
class VideoFileMedia(_VideoMedia):
    """
    A video media that is read from a video
    file and can be subclipped to a specific
    time range.
    """

    @property
    def copy(
        self
    ) -> 'VideoFileMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoFileMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy

    @property
    def ticks_per_frame(
        self
    ) -> int:
        """
        The number of ticks per video frame. A
        tick is the minimum amount of time and
        is the way 'pts' is measured, in ticks.

        This means that the 'pts' value will
        be increased this amount from one video
        frame to the next one.

        How we obtain it:
        - `(1 / fps) / time_base`
        """
        return self.source.ticks_per_frame
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the video, that can be
        shorter than the video file source
        duration if the user requested for it.

        The formula:
        - `self.end - self.start`
        """
        return self.end - self.start
    
    @property
    def number_of_frames(
        self
    ) -> Union[int, None]:
        """
        The number of frames of the video.
        """
        return self.source.number_of_frames
    
    @property
    def fps(
        self
    ) -> Union[Fraction, None]:
        """
        The frames per second of the video.
        """
        return self.source.fps
    
    @property
    def audio_fps(
        self
    ) -> Union[int, None]:
        """
        The frames per second of the audio.
        """
        return self.source.audio_fps
    
    @property
    def codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the video codec.
        """
        return self.source.codec_name
    
    @property
    def audio_codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the audio codec.
        """
        return self.source.audio_codec_name
    
    @property
    def audio_layout(
        self
    ) -> Union[str, None]:
        """
        The audio layout.
        """
        return self.source.audio_layout.name
    
    @property
    def audio_format(
        self
    ) -> Union[str, None]:
        """
        The audio format.
        """
        return self.source.audio_format.name
    
    @property
    def pixel_format(
        self
    ) -> Union[str, None]:
        """
        The pixel format.
        """
        return self.source.pixel_format

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the video frames expressed 
        like (width, height).
        """
        return self.source.size
    
    @property
    def width(
        self
    ) -> int:
        """
        The width of the video frames in pixels.
        """
        return self.size[0]
    
    @property
    def height(
        self
    ) -> int:
        """
        The height of the video frames in pixels.
        """
        return self.size[1]
    
    @property
    def time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the video.
        """
        return self.source.time_base
    
    @property
    def audio_time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the audio.
        """
        return self.source.audio_time_base

    def __init__(
        self,
        filename: str,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
        size: Union[tuple[int, int], None] = None,
        # These are ffmpeg filters
        video_filters: list[GraphFilter] = [],
        audio_filters: list[GraphFilter] = []
    ):
        # We first create the source with no
        # filters just to be able to know the
        # original size and properties
        super().__init__(
            source = VideoFileSource(
                filename = filename,
                video_filters = [],
                audio_filters = []
            ),
            start = start,
            end = end
        )
        
        size: tuple[int, int] = (
            self.size
            if size is None else
            size
        )

        self.output_size: tuple[int, int] = None
        """
        The size we expect to be used as the
        output size. It can be different from
        the original vide source size.
        """

        # Now we can set the filters properly
        self.set_video_filters(video_filters)
        self.set_audio_filters(audio_filters)
        self.set_output_size(size)

    def set_output_size(
        self,
        output_size: tuple[int, int]
    ) -> 'VideoFileMedia':
        """
        Set the desired size for the output, that
        means the video frames when rendering. This
        method will also update the scaling filter.
        """
        self.output_size = output_size
        
        if self.size != self.output_size:
            # We need to apply a resize (scale) pyav
            # filter to obtain the desired output size
            self._set_scale_filter(self.output_size)

    # TODO: Do we need this in the source instead (?)
    def _set_scale_filter(
        self,
        size: tuple[int, int]
    ) -> 'VideoFileMedia':
        """
        Set the scale 'filter' provided replacing
        any other previous scale filter set in this
        class.
        """
        # Remove previous 'scale' filter if existing
        self.set_video_filters(
            [
                video_filter
                for video_filter in self.source._video_filters
                if video_filter.name != 'scale'
            ]
        )

        self.add_video_filter(GraphFilters.video.scale(size[0], size[1]))

        return self

    def add_video_filter(
        self,
        filter: GraphFilter
    ) -> 'VideoFileMedia':
        """
        Add a video filter to the list of filters
        to apply.
        """
        self.source.add_video_filter(filter)

        return self
    
    def set_video_filters(
        self,
        filters: list[GraphFilter] = []
    ) -> 'VideoFileMedia':
        """
        Set the 'filters' provided as the new video
        filters, replacing the previous one if 
        existing.
        """
        # TODO: What if 'audio' filters coming (?)
        self.source.set_video_filters(filters)

        return self
    
    def add_audio_filter(
        self,
        filter: GraphFilter
    ) -> 'VideoFileMedia':
        """
        Add an audio filter to the list of filters
        to apply.
        """
        self.source.add_audio_filter(filter)

        return self
    
    def set_audio_filters(
        self,
        filters: list[GraphFilter] = []
    ) -> 'VideoFileMedia':
        """
        Set the 'filters' provided as the new audio
        filters, replacing the previous one if 
        existing.
        """
        # TODO: What if 'video' filters coming (?)
        self.source.set_audio_filters(filters)

        return self

class VideoImageMedia(_VideoMedia):
    """
    A video media that is made by an static
    image file.
    """

    @property
    def copy(
        self
    ) -> 'VideoImageMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoImageMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy

    @property
    def filename(
        self
    ) -> str:
        """
        The filename of the original image.
        """
        return self.source.filename

    # TODO: Maybe rename this property (?)
    @property
    def do_include_alpha(
        self
    ) -> bool:
        """
        The internal flag to indicate if we
        want to consider the alpha channel or
        not.
        """
        return self.source._do_include_alpha
    
    def __init__(
        self,
        filename: str,
        duration: Union[int, float, Fraction],
        do_include_alpha: bool = True,
        size: Union[tuple[int, int], None] = None,
        # Need this to generate the silent audio frames
        audio_fps: int = 44_100,
        audio_samples_per_frame: int = 1024,
    ):
        """
        TODO: Maybe force size provided as parameter (?)
        """
        # We need to dynamically set the frame format
        # to accept (or not) the alpha channel
        frame_format = (
            'rgba'
            if do_include_alpha else
            'rgb24'
        )

        super().__init__(
            source = VideoImageSource(
                filename = filename,
                do_include_alpha = do_include_alpha,
                frame_format = frame_format,
                audio_fps = audio_fps,
                audio_samples_per_frame = audio_samples_per_frame
            ),
            start = 0,
            end = duration
        )

        self.output_size: tuple[int, int] = None
        """
        The size we expect to be used as the
        output size. It can be different from
        the original vide source size.
        """

        self.set_output_size(
            self.size
            if size is None else
            size
        )

    def set_output_size(
        self,
        output_size: tuple[int, int]
    ) -> 'VideoFileMedia':
        """
        Set the desired size for the output, that
        means the video frames when rendering. This
        method will also update the scaling filter.
        """
        self.output_size = output_size

class VideoColorMedia(_VideoMedia):
    """
    A video media that is made with a static
    uniform color frame.
    """

    @property
    def copy(
        self
    ) -> 'VideoColorMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoColorMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy
    
    @property
    def color(
        self
    ) -> tuple[int, int, int]:
        """
        The color that will be used to make the
        frame that will be played its whole
        duration.
        """
        return self.source._color
    
    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the media frame.
        """
        return self.source.size

    def __init__(
        self,
        color: Color,
        duration: Union[int, float, Fraction],
        size: tuple[int, int] = (1920, 1080),
        transparency: Union[float, None] = 0.0,
        # TODO: Should we accept 'frame_format' (?)
        #frame_format: str = 'rgba',
        # Need this to generate the silent audio frames
        audio_fps: int = 44_100,
        audio_samples_per_frame: int = 1024,
    ):
        # TODO: What about the 'transparency' if they
        # provide a 'color' also with 'transparency' (?)
        color = Color.parse(color)

        frame_format = (
            'rgba'
            if transparency is not None else
            'rgb24'
        )

        super().__init__(
            source = VideoColorSource(
                color = color.rgb_not_normalized,
                size = size,
                frame_format = frame_format,
                transparency = transparency,
                audio_fps = audio_fps,
                audio_samples_per_frame = audio_samples_per_frame
            ),
            start = 0,
            end = duration
        )

        self.output_size: tuple[int, int] = None
        """
        The size we expect to be used as the
        output size. It can be different from
        the original vide source size.
        """

        self.set_output_size(self.size)

    def set_output_size(
        self,
        output_size: tuple[int, int]
    ) -> 'VideoFileMedia':
        """
        Set the desired size for the output, that
        means the video frames when rendering. This
        method will also update the scaling filter.
        """
        self.output_size = output_size

# TODO: Create 'VideoNumpyMedia'


from yta_editor.media.abstract import _Media
from yta_editor.sources.abstract import _AudioSource
from yta_editor.sources.audio import AudioFileSource, AudioNumpySource
from yta_editor.decorators import with_t_adjusted_to_media
from yta_editor.utils.effects import apply_audio_effects_to_frame_at_t
from yta_video_pyav.reader.filter.dataclass import GraphFilter
from yta_video_opengl.nodes import TimedNode
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


class _AudioMedia(_Media):
    """
    Abstract class to be inherited by any
    media element.

    The media element is an element that
    includes a source and a 'start' and
    'end' values to be able to subclip that
    media source and use only the part we
    want to use.
    """

    def __init__(
        self,
        source: _AudioSource,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
    ):
        super().__init__(
            source = source,
            start = start,
            end = end
        )

    def add_effect(
        self,
        effect: TimedNode
    ) -> '_AudioMedia':
        """
        Add the provided 'effect' to the audio.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, 'TimedNode')

        if not effect.is_audio_node:
            raise Exception('The provided "effect" is not an audio effect.')

        self._effects.add_effect(effect)
        
        return self
    
    @with_t_adjusted_to_media
    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction]
    ):
        """
        Get the sequence of audio frames for a 
        given video 't' time moment, using the
        audio cache system.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        print(f'Getting audio frames from {str(float(t + self.start))} that is actually {str(float(t))}')
        for frame in self.source.get_audio_frames_at_t(t, video_fps):
            yield apply_audio_effects_to_frame_at_t(
                effects_stack = self._effects,
                frame = frame,
                t = t
            )

class AudioFileMedia(_AudioMedia):
    """
    An audio media that is read from an audio
    file and can be subclipped to a specific
    time range.
    """

    @property
    def copy(
        self
    ) -> 'AudioFileMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = AudioFileMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy

    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the video.
        """
        return self.end - self.start
    
    @property
    def audio_fps(
        self
    ) -> Union[int, None]:
        """
        The frames per second of the audio.
        """
        return self.source.audio_fps
    
    @property
    def audio_codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the audio codec.
        """
        return self.source.audio_codec_name
    
    @property
    def audio_layout(
        self
    ) -> Union[str, None]:
        """
        The audio layout.
        """
        return self.source.audio_layout.name
    
    @property
    def audio_format(
        self
    ) -> Union[str, None]:
        """
        The audio format.
        """
        return self.source.audio_format.name
    
    @property
    def audio_time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the audio.
        """
        return self.source.audio_time_base
    
    def __init__(
        self,
        filename: str,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
        # These are ffmpeg filters
        audio_filters: list[GraphFilter] = []
    ):
        super().__init__(
            source = AudioFileSource(
                filename = filename,
                audio_filters = audio_filters
            ),
            start = start,
            end = end
        )

    def add_audio_filter(
        self,
        filter: GraphFilter
    ) -> 'AudioFileMedia':
        """
        Add an audio filter to the list of filters
        to apply.
        """
        self.source.add_audio_filter(filter)

        return self

# TODO: Create 'AudioNumpyMedia'


from yta_video_opengl.effects import EffectsStack
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union
from abc import ABC, abstractmethod


class _Media(ABC):
    """
    Abstract class to be inherited by any
    media element.

    The media element is an element that
    includes a source and a 'start' and
    'end' values to be able to subclip that
    media source and use only the part we
    want to use.
    """

    @property
    @abstractmethod
    def copy(
        self
    ) -> '_Media':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        pass

    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the media, that can be
        shorter than the source duration if the
        user requested it.

        The formula:
        - `self.end - self.start`
        """
        return self.end - self.start
    
    @property
    def time_remaining_until_start(
        self
    ) -> Fraction:
        """
        The time remaining at the begining of this
        Media source due to the 'start' provided to
        this instance.

        The source will always start on 0, so the
        actual start we use in this Media (maybe 2)
        is also the remaining time we didn't use at
        the begining (the first 2 seconds in this
        same example).

        This value can be useful for transitions.

        The formula:
        - `self.start`
        """
        return self.start
    
    @property
    def time_remaining_since_end(
        self
    ) -> Fraction:
        """
        The time remaining at the end of this Media
        source due to the 'end' provided to this
        instance.

        The source has its own duration, but maybe
        the 'end' we use is 2 seconds before that
        actual end moment, so the time remaining
        (that we didn't use) at the end of the source
        is the actual source duration minus the 'end'
        we provided here (2 seconds in this example I
        said).

        This value can be useful for transitions.

        The formula:
        - `self.source.duration - self.end`
        """
        return self.source.duration - self.end

    def __init__(
        self,
        source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'],
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
    ):
        self.source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'] = source
        """
        The source of this media element that
        is the entity from which we can obtain
        the frames.
        """
        self._effects: EffectsStack = EffectsStack()
        """
        The effects we want to apply on the
        media.
        """
        self.start: Fraction
        """
        The time moment 't' in which the media
        should start being played/displayed.
        """
        self.end: Union[Fraction, None]
        """
        The time moment 't' in which the media
        should end being played/displayed.
        """
        
        # Set 'start' and 'end'
        self.set_time_range(start, end)

    def set_time_range(
        self,
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction, None] = None,
    ) -> '_Media':
        """
        Set the media 'start' and 'end' time
        moments range from the original source
        that will be played/displayed.

        - If `end = None`, the source duration
        (if available) will be set.
        - If `end > source.duration` it will be
        replaced by the source duration value.
        """
        ParameterValidator.validate_mandatory_positive_number('start', start, do_include_zero = True)
        ParameterValidator.validate_positive_number('end', end, do_include_zero = False)

        self.start: Fraction = Fraction(start)
        self.end: Union[Fraction, None] = Fraction(
            # TODO: Is this 'end' ok (?)
            self.source.duration
            if (
                end is None or
                (
                    self.source.duration is not None and
                    end > self.source.duration
                )
            ) else
            end
        )

        # If the source has a duration, the 'start'
        # and 'end' must be valid
        if self.source.duration is not None:
            if (
                self.start >= self.source.duration and
                self.end >= self.source.duration
            ):
                raise Exception(f'The provided "start" and "end" are invalid values considering the real media duration of {str(float(self.source.duration))}s')
            
            if self.end <= self.start:
                raise Exception('The "end" value cannot be equal or smaller than the "start" value.')
            
            self.end = (
                self.source.duration
                if self.end > self.source.duration else
                self.end
            )

        return self
    
    # TODO: This method has been created
    # to be inherited by the other classes
    # and being able to copy the instance
    # properly by using the same 'source'
    # reference and creating not a new one
    @classmethod
    def _init_with_source(
        cls,
        source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'],
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None
    ):
        """
        *For internal use only*

        Alternative '__init__' to create the
        instance from the 'source' directly. This
        method must be called by the specific
        implementations of this abstract class
        to be able to instantiate them directly
        from the 'source' to make copies.

        We created this method to avoid generating
        a new 'source' instance but preserving the
        same reference.
        """
        # Create new instance skipping '__init__'
        instance = cls.__new__(cls)
        super(cls, instance).__init__(
            source = source,
            start = start,
            end = end
        )

        return instance

